# SPDX-FileCopyrightText: 2019 Dave Astels for Adafruit Industries
#
# SPDX-License-Identifier: MIT

"""
`adafruit_debouncer`
====================================================

Debounces an arbitrary predicate function (typically created as a lambda) of 0
arguments.  Since a very common use is debouncing a digital input pin, the
initializer accepts a DigitalInOut object instead of a lambda.

* Author(s): Dave Astels

Implementation Notes
--------------------

**Software and Dependencies:**

* Adafruit CircuitPython firmware for the supported boards:
  https://github.com/adafruit/circuitpython/releases
* Adafruit Ticks library:
  https://github.com/adafruit/Adafruit_CircuitPython_Ticks
"""

# imports

__version__ = "2.0.5"
__repo__ = "https://github.com/adafruit/Adafruit_CircuitPython_Debouncer.git"

from micropython import const
from adafruit_ticks import ticks_ms, ticks_diff

try:
    from typing import Callable, Optional, Union
    from circuitpython_typing.io import ROValueIO
except ImportError:
    pass

_DEBOUNCED_STATE: int = const(0x01)
_UNSTABLE_STATE: int = const(0x02)
_CHANGED_STATE: int = const(0x04)

_TICKS_PER_SEC: int = const(1000)


class Debouncer:
    """Debounce an input pin or an arbitrary predicate"""

    def __init__(
        self,
        io_or_predicate: Union[ROValueIO, Callable[[], bool]],
        interval: float = 0.010,
    ) -> None:
        """Make an instance.
        :param DigitalInOut/function io_or_predicate: the DigitalIO or
                                                      function that returns a boolean to debounce
        :param float interval: bounce threshold in seconds (default is 0.010, i.e. 10 milliseconds)
        """
        self.state = 0x00
        if hasattr(io_or_predicate, "value"):
            self.function = lambda: io_or_predicate.value
        else:
            self.function = io_or_predicate
        if self.function():
            self._set_state(_DEBOUNCED_STATE | _UNSTABLE_STATE)
        self._last_bounce_ticks = 0
        self._last_duration_ticks = 0
        self._state_changed_ticks = 0

        # Could use the .interval setter, but pylint prefers that we explicitly
        # set the real underlying attribute:
        self._interval_ticks = int(interval * _TICKS_PER_SEC)

    def _set_state(self, bits: int) -> None:
        self.state |= bits

    def _unset_state(self, bits: int) -> None:
        self.state &= ~bits

    def _toggle_state(self, bits: int) -> None:
        self.state ^= bits

    def _get_state(self, bits: int) -> bool:
        return (self.state & bits) != 0

    def update(self, new_state: Optional[int] = None) -> None:
        """Update the debouncer state. MUST be called frequently"""
        now_ticks = ticks_ms()
        self._unset_state(_CHANGED_STATE)
        if new_state is None:
            current_state = self.function()
        else:
            current_state = bool(new_state)
        if current_state != self._get_state(_UNSTABLE_STATE):
            self._last_bounce_ticks = now_ticks
            self._toggle_state(_UNSTABLE_STATE)
        else:
            if ticks_diff(now_ticks, self._last_bounce_ticks) >= self._interval_ticks:
                if current_state != self._get_state(_DEBOUNCED_STATE):
                    self._last_bounce_ticks = now_ticks
                    self._toggle_state(_DEBOUNCED_STATE)
                    self._set_state(_CHANGED_STATE)
                    self._last_duration_ticks = ticks_diff(
                        now_ticks, self._state_changed_ticks
                    )
                    self._state_changed_ticks = now_ticks

    @property
    def interval(self) -> float:
        """The debounce delay, in seconds"""
        return self._interval_ticks / _TICKS_PER_SEC

    @interval.setter
    def interval(self, new_interval_s: float) -> None:
        self._interval_ticks = new_interval_s * _TICKS_PER_SEC

    @property
    def value(self) -> bool:
        """Return the current debounced value."""
        return self._get_state(_DEBOUNCED_STATE)

    @property
    def rose(self) -> bool:
        """Return whether the debounced value went from low to high at the most recent update."""
        return self._get_state(_DEBOUNCED_STATE) and self._get_state(_CHANGED_STATE)

    @property
    def fell(self) -> bool:
        """Return whether the debounced value went from high to low at the most recent update."""
        return (not self._get_state(_DEBOUNCED_STATE)) and self._get_state(
            _CHANGED_STATE
        )

    @property
    def last_duration(self) -> float:
        """Return the number of seconds the state was stable prior to the most recent transition."""
        return self._last_duration_ticks / _TICKS_PER_SEC

    @property
    def current_duration(self) -> float:
        """Return the number of seconds since the most recent transition."""
        return ticks_diff(ticks_ms(), self._state_changed_ticks) / _TICKS_PER_SEC


class Button(Debouncer):
    """
    Debouncer for buttons. Reports ``pressed`` and ``released`` for the button state.
    Counts multiple short presses, allowing to detect double clicks, triple clicks, etc.
    Reports long presses separately. A long press can immediately follow multiple clicks,
    in which case the long click will be reported in the same update as the short clicks.

    :param DigitalInOut/function pin: the DigitalIO or function to debounce.
    :param int short_duration_ms: the maximum length of a short press in milliseconds.
    :param int long_duration_ms: the minimum length of a long press in milliseconds.
    :param bool value_when_pressed: the value of the predicate when the button is
                                    pressed. Defaults to False (for pull up buttons).
    """

    def __init__(
        self,
        pin: Union[ROValueIO, Callable[[], bool]],
        short_duration_ms: int = 200,
        long_duration_ms: int = 500,
        value_when_pressed: bool = False,
        **kwargs
    ) -> None:
        self.short_duration_ms = short_duration_ms
        self.long_duration_ms = long_duration_ms
        self.value_when_pressed = value_when_pressed
        self.last_change_ms = ticks_ms()
        self.short_counter = 0
        self.short_to_show = 0
        self.long_registered = False
        self.long_to_show = False
        super().__init__(pin, **kwargs)

    @property
    def pressed(self) -> bool:
        """Return whether the button was pressed or not at the last update."""
        return (self.value_when_pressed and self.rose) or (
            not self.value_when_pressed and self.fell
        )

    @property
    def released(self) -> bool:
        """Return whether the button was release or not at the last update."""
        return (self.value_when_pressed and self.fell) or (
            not self.value_when_pressed and self.rose
        )

    def update(self, new_state: Optional[int] = None) -> None:
        super().update(new_state)
        if self.pressed:
            self.last_change_ms = ticks_ms()
            self.short_counter = self.short_counter + 1
        elif self.released:
            self.last_change_ms = ticks_ms()
            if self.long_registered:
                self.long_registered = False
        else:
            duration = ticks_diff(ticks_ms(), self.last_change_ms)
            if (
                not self.long_registered
                and self.value == self.value_when_pressed
                and duration > self.long_duration_ms
            ):
                self.long_registered = True
                self.long_to_show = True
                self.short_to_show = self.short_counter - 1
                self.short_counter = 0
            elif (
                self.short_counter > 0
                and self.value != self.value_when_pressed
                and duration > self.short_duration_ms
            ):
                self.short_to_show = self.short_counter
                self.short_counter = 0
            else:
                self.long_to_show = False
                self.short_to_show = 0

    @property
    def short_count(self) -> int:
        """Return the number of short press if a series of short presses has
        ended at the last update."""
        return self.short_to_show

    @property
    def long_press(self) -> bool:
        """Return whether a long press has occured at the last update."""
        return self.long_to_show



# SPDX-FileCopyrightText: 2021 ladyada for Adafruit Industries
#
# SPDX-License-Identifier: MIT

"""
How to use this test file:

Copy adafruit_debouncer's dependencies to lib/ on your circuitpython device.
Copy adafruit_debouncer.py to / on the device
Copy this tests.py file to /main.py on the device
Connect to the serial terminal (e.g. sudo screen /dev/ttyACM0 115200)
Press Ctrl-D, if needed to start the tests running
"""
import sys
import time
import adafruit_debouncer


def _true():
    return True


def _false():
    return False


def assertEqual(var_a, var_b):  # pylint: disable=invalid-name
    assert var_a == var_b, "Want %r, got %r" % (var_a, var_b)


def test_back_and_forth():
    # Start false
    debouncer = adafruit_debouncer.Debouncer(_false)
    assertEqual(debouncer.value, False)

    # Set the raw state to true, update, and make sure the debounced
    # state has not changed yet:
    debouncer.function = _true
    debouncer.update()
    assertEqual(debouncer.value, False)
    assert not debouncer.last_duration, "There was no previous interval??"

    # Sleep longer than the debounce interval, so state can change:
    time.sleep(0.02)
    debouncer.update()
    assert debouncer.last_duration  # is actually duration between powerup and now
    assertEqual(debouncer.value, True)
    assertEqual(debouncer.rose, True)
    assertEqual(debouncer.fell, False)
    # Duration since last change has only been long enough to run these
    # asserts, which should be well under 1/10 second
    assert debouncer.current_duration < 0.1, (
        "Unit error? %d" % debouncer.current_duration
    )

    # Set raw state back to false, make sure it's not instantly reflected,
    # then wait and make sure it IS reflected after the interval has passed.
    debouncer.function = _false
    debouncer.update()
    assertEqual(debouncer.value, True)
    assertEqual(debouncer.fell, False)
    assertEqual(debouncer.rose, False)
    time.sleep(0.02)
    assert 0.019 < debouncer.current_duration <= 1, (
        "Unit error? sleep .02 -> duration %d" % debouncer.current_duration
    )
    debouncer.update()
    assertEqual(debouncer.value, False)
    assertEqual(debouncer.rose, False)
    assertEqual(debouncer.fell, True)

    assert 0 < debouncer.current_duration <= 0.1, (
        "Unit error? time to run asserts %d" % debouncer.current_duration
    )
    assert 0 < debouncer.last_duration < 0.1, (
        "Unit error? Last dur should be ~.02, is %d" % debouncer.last_duration
    )


def test_interval_is_the_same():
    debouncer = adafruit_debouncer.Debouncer(_false, interval=0.25)
    assertEqual(debouncer.value, False)
    debouncer.update()
    debouncer.function = _true
    debouncer.update()

    time.sleep(0.1)  # longer than default interval
    debouncer.update()
    assertEqual(debouncer.value, False)

    time.sleep(0.2)  # 0.1 + 0.2 > 0.25
    debouncer.update()
    assertEqual(debouncer.value, True)
    assertEqual(debouncer.rose, True)
    assertEqual(debouncer.interval, 0.25)


def test_setting_interval():
    # Check that setting the interval does change the time the debouncer waits
    debouncer = adafruit_debouncer.Debouncer(_false, interval=0.01)
    debouncer.update()

    # set the interval to a longer time, sleep for a time between
    # the two interval settings, and assert that the value hasn't changed.

    debouncer.function = _true
    debouncer.interval = 0.2
    debouncer.update()
    assert debouncer.interval - 0.2 < 0.00001, "interval is not consistent"
    time.sleep(0.11)
    debouncer.update()

    assertEqual(debouncer.value, False)
    assertEqual(debouncer.rose, False)
    assertEqual(debouncer.fell, False)

    # and then once the whole time has passed make sure it did change
    time.sleep(0.11)
    debouncer.update()
    assertEqual(debouncer.value, True)
    assertEqual(debouncer.rose, True)
    assertEqual(debouncer.fell, False)


def run():
    passes = 0
    fails = 0
    for name, test in locals().items():
        if name.startswith("test_") and callable(test):
            try:
                print()
                print(name)
                test()
                print("PASS")
                passes += 1
            except Exception as err:  # pylint: disable=broad-except
                sys.print_exception(err)  # pylint: disable=no-member
                print("FAIL")
                fails += 1

    print(passes, "passed,", fails, "failed")
    if passes and not fails:
        print(
            r"""
 ________
< YATTA! >
 --------
        \   ^__^
         \  (oo)\_______
            (__)\       )\/\
                ||----w |
                ||     ||"""
        )


if __name__ == "__main__":
    run()



# SPDX-FileCopyrightText: 2019 Dave Astels for Adafruit Industries
# SPDX-License-Identifier: MIT

# pylint: disable=invalid-name

# Wait for a falling transition on the clock (signal 1)
# When it's seen, display the values of signal 2-5
# Demonstrates debouncing a lambda predicate

import time
from adafruit_crickit import crickit
from adafruit_debouncer import Debouncer

ss = crickit.seesaw


def make_crikit_signal_debouncer(pin):
    """Return a lambda to read the specified pin"""
    ss.pin_mode(pin, ss.INPUT_PULLUP)
    return Debouncer(lambda: ss.digital_read(pin))


# Two buttons are pullups, connect to ground to activate
clock = make_crikit_signal_debouncer(crickit.SIGNAL1)
signal_2 = make_crikit_signal_debouncer(crickit.SIGNAL2)
signal_3 = make_crikit_signal_debouncer(crickit.SIGNAL3)
signal_4 = make_crikit_signal_debouncer(crickit.SIGNAL4)
signal_5 = make_crikit_signal_debouncer(crickit.SIGNAL5)

while True:
    clock.update()
    signal_2.update()
    signal_3.update()
    signal_4.update()
    signal_5.update()

    if clock.fell:
        print(
            "%u %u %u %u"
            % (signal_2.value, signal_3.value, signal_4.value, signal_5.value)
        )

    time.sleep(0.1)



# SPDX-FileCopyrightText: 2021 ladyada for Adafruit Industries
# SPDX-License-Identifier: MIT

import time
import board
from digitalio import DigitalInOut, Direction, Pull
from adafruit_debouncer import Debouncer

button = DigitalInOut(board.D4)
button.direction = Direction.INPUT
button.pull = Pull.UP
switch = Debouncer(button)

while True:
    switch.update()
    if switch.fell:
        print("pressed")
        print("was released for ", switch.last_duration)
    elif switch.rose:
        print("released")
        print("was pressed for ", switch.last_duration)
    else:
        print("Stable for ", switch.current_duration)
    time.sleep(0.1)



# SPDX-FileCopyrightText: 2021 ladyada for Adafruit Industries
# SPDX-License-Identifier: MIT

"""
This example shows how to use the debouncer library on the signals coming from
a cap-sense pin with touchio.
"""
import time
import board
import touchio
from adafruit_debouncer import Debouncer

touch_pad = board.A1
touch = touchio.TouchIn(touch_pad)
touch_debounced = Debouncer(touch)

while True:
    touch_debounced.update()
    if touch_debounced.fell:
        print("Just released")
    if touch_debounced.rose:
        print("Just pressed")
    if touch_debounced.value:
        print("touching")
    else:
        # print('not touching')
        pass
    time.sleep(0.05)



# SPDX-FileCopyrightText: 2022 david gauchard
# SPDX-License-Identifier: MIT

import board
import digitalio
from adafruit_debouncer import Button

# This example shows how to count short clicks or detect a long press

pin = digitalio.DigitalInOut(board.D12)
pin.direction = digitalio.Direction.INPUT
pin.pull = digitalio.Pull.UP
switch = Button(pin)

while True:
    switch.update()
    if switch.long_press:
        print("Long Press")
    if switch.short_count != 0:
        print("Short Press Count =", switch.short_count)
    if switch.long_press and switch.short_count == 1:
        print("That's a long double press !")



# SPDX-FileCopyrightText: 2019 Dave Astels for Adafruit Industries
# SPDX-License-Identifier: MIT

# pylint: disable=invalid-name

import board
import digitalio
from adafruit_debouncer import Debouncer

pin = digitalio.DigitalInOut(board.D12)
pin.direction = digitalio.Direction.INPUT
pin.pull = digitalio.Pull.UP
switch = Debouncer(pin)

while True:
    switch.update()
    if switch.fell:
        print("Just pressed")
    if switch.rose:
        print("Just released")
    if switch.value:
        print("not pressed")
    else:
        print("pressed")
